package cn.edu.zjvtit.cloudstorage.apiservice.controller;

import cn.edu.zjvtit.cloudstorage.apiservice.Orm;
import cn.edu.zjvtit.cloudstorage.apiservice.entity.Entitylnl;
import cn.edu.zjvtit.cloudstorage.apiservice.rabbitmq.topic.Sender;
import cn.edu.zjvtit.cloudstorage.apiservice.store.DataServerStore;
import com.rabbitmq.client.Channel;
import org.slf4j.Logger;
import org.springframework.amqp.rabbit.core.RabbitTemplate;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpMethod;
import org.springframework.http.ResponseEntity;
import org.springframework.util.DigestUtils;
import org.springframework.web.bind.annotation.*;
import cn.edu.zjvtit.cloudstorage.apiservice.result.Result;
import org.springframework.web.client.RestTemplate;

import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.OutputStream;
import java.util.HashMap;
import java.util.Map;

@RestController
public class ObjectsController {
    @Autowired
    private DataServerStore dataServerStore;
    @Autowired
    private Sender sender;

    @Autowired
    private Logger logger;
    @Autowired
    private RestTemplate restTemplate;
    @Autowired
    private RabbitTemplate rabbitTemplate;
    @Autowired
    private Orm orm;
    @PutMapping("/objects/{filename}")
    public String putObject(@RequestBody byte[] requestBody, @RequestHeader Map<String, String> headers, @PathVariable("filename") String filename) throws IOException {
        String server = this.dataServerStore.selectRandomServer();
        logger.info("put object to server: " + server);
        if (server == null) {
            return new Result("no server available", "10000").toString();
        }
        String md5 = DigestUtils.md5DigestAsHex(requestBody);
        Entitylnl object = this.orm.findByName(filename);
        if (object == null) {
            object = new Entitylnl();
            object.setName(filename);
            object.setHash(md5);
            object.setVersion(1);
            object.setSize(requestBody.length);
        } else {
            if (md5.equals(object.getHash())){
                return new Result("object already exists", "10001").toString();
            }
            Integer version = object.getVersion()+1;
            object = new Entitylnl();
            object.setName(filename);
            object.setHash(md5);
            object.setVersion(version);
            object.setSize(requestBody.length);
        }
        this.orm.save(object);
        String url = "http://" + server + "/objects/" + filename;
        // 发送post请求
        HttpEntity<byte[]> request = new HttpEntity<>(requestBody);
        this.restTemplate.exchange(url, HttpMethod.PUT, request, String.class);
        return (new Result("success", "10000")).toString();
    }


    @GetMapping("/objects/{filename}")
    public String getObject(HttpServletResponse response, @PathVariable("filename") String filename) throws IOException {
        // 查询哪台服务器有这个文件
        // 读取文件
        Entitylnl obj;
        int version = 0;
        if (version == 0) {
            obj = this.orm.findFirstByNameOrderByVersionDesc(filename);
        } else {
            obj = this.orm.findByVersionAndName(version, filename);
        }
        String queueName = filename;
        Channel channel = rabbitTemplate.getConnectionFactory().createConnection().createChannel(false);
        channel.queueDeclare(queueName, false, false, false, null);
        channel.queueBind(queueName, "exchange", "");
        sender.send(queueName);
        try {
            Object msg = rabbitTemplate.receiveAndConvert(queueName, 200);
            assert msg != null;
            logger.info(msg.toString());
            String url = "http://" + msg + "/objects/" + filename;

            ResponseEntity<String> res = this.restTemplate.getForEntity(url,String.class);
            response.reset();
            response.setContentType("application/octet-stream");
            response.setCharacterEncoding("utf-8");
            response.setContentLength((int) res.getHeaders().getContentLength());
            response.setHeader("Content-Disposition", "attachment;filename=" + filename );
            OutputStream os = response.getOutputStream();
            os.write(res.getBody().getBytes());
            os.close();
            return "";
        } catch (Exception e) {
            logger.info("locate object error", e);
            return new Result("file not exists", "10000").toString();
        }
    }

    @GetMapping("/locate/{filename}")
    public String locateObject(@PathVariable("filename") String filename) throws IOException {
        // 查询哪台服务器有这个文件
        String queueName = filename;
        Channel channel = rabbitTemplate.getConnectionFactory().createConnection().createChannel(false);
        channel.queueDeclare(queueName, false, false, false, null);
        channel.queueBind(queueName, "exchange", "");
        sender.send(queueName);
        try {
            Object msg = rabbitTemplate.receiveAndConvert(queueName, 200);
            assert msg != null;
            logger.info(msg.toString());
            return new Result("file in server:" + msg.toString(), "10000").toString();
        } catch (Exception e) {
            logger.info("locate object error", e);
            return new Result("file not exists", "10000").toString();
        }

    }
    @GetMapping("/versions/{filename}")
    public String getVersions(@PathVariable("filename") String filename,
                              @RequestParam("offset") int offset, @RequestParam("limit") int limit) {
        if (offset <= 0) {
            offset = 0;
        }
        if (limit <= 0) {
            limit = 10;
        }
        Page<Entitylnl> result =
                this.orm.findAllByNameOrderByVersionDesc(filename,
                        PageRequest.of(offset, limit));
        Map<String, Object> map = new HashMap<>();
        map.put("total", result.getTotalElements());
        map.put("rows", result.getContent());
        return new Result("file not exists", "10000").withData(map).toString();
    }
    @GetMapping("/versions")
    public String getVersionList(@RequestParam("offset") int offset,
                                 @RequestParam("limit") int limit) {
        if (offset <= 0) {
            offset = 0;
        }
        if (limit <= 0) {
            limit = 10;
        }
        Page<Entitylnl> result =
                this.orm.findAll(PageRequest.of(offset, limit));
        Map<String, Object> map = new HashMap<>();
        map.put("total", result.getTotalElements());
        map.put("rows", result.getContent());
        return new Result("file not exists", "10000").withData(map).toString();
    }

}
